//
// Copyright 1996, Thomas Andrews (thomaso@best.com)
// Free for non-commercial use
//

// Expects an array of CARDS*HANDS integers from 0 to HANDS-1,
// representing where each card in the deck belongs
//
//  No sanity checking is done on this array
///////
template <int CARDS, int HANDS> 
tDeal<CARDS, HANDS>::tDeal(const int *deal_array) {
  int card, index, value;
  for (card=0; card<HANDS*CARDS; card++) {
    whom[card]=deal_array[card];
  }
}

// As with above constructor, only expects an array of chars, from
// '0' to '0'+HANDS-1.
//
// No sanity checking is done on this data
///////
template <int CARDS, int HANDS>
tDeal<CARDS, HANDS>::tDeal(const char *deal_array) {
  int card;
  for (card=0; card<HANDS*CARDS; card++) {
    whom[card]=(deal_array[card]-'0');
  }
}

// The algorithm is described elsewhere.
//
// Rough outline (using normal deck parameters - HANDS=4, CARDS=13):
//	
//	Find the subset of {0,1,...,52-1} which corresponds to the cards
//         held by north.
//
//      Find the 13-element subset of {0,...,39-1} which corresponds to
//         the cards which remain which go to east.
//
//      Find the 13-element subset of {0,...,26-1} which corresponds to
//         the which remain and go to east.
//
//      Encode these 13-sets with the "squashed order index."
//
//      Put the three resulting numbers together into one giant number.
//
// See the web page: http://www.best.com/~thomaso/decode/algorithm.html
// for a more verbose description.
//
////////
template <int CARDS, int HANDS>
tDeal<CARDS, HANDS>::operator Integer() {
   Integer result=0;
   int hand;
   for (hand=0; hand<HANDS-1; hand++) {
     int card, index, index2;
     int seq_array[CARDS];
     for (card=0, index=0,index2=0; card<HANDS*CARDS;  card++) {
	if (whom[card]>=hand) {
 	  if (whom[card]==hand) { seq_array[index2++]=index; }
          index++;
        }
     }
     result *= Choose((HANDS-hand)*CARDS,CARDS);
     result += Integer(IncrSequence(CARDS,seq_array));
   }
   return result;
}

// The decoding routine.
// The algorithm is described elsewhere.
template <int CARDS, int HANDS>
tDeal<CARDS, HANDS>::tDeal(Integer val) {
    Integer Index, quotient;
    IncrSequence *Seq[HANDS-1];

    int i=0;
    while (i<HANDS-2) {
        // Would rather say
	//      divide(val, Choose((i+2)*CARDS,CARDS), val, Index);
        // here, but Integer "divide" does not like using the
	// val variable both as input and output...
	divide(val, Choose((i+2)*CARDS,CARDS), quotient, Index);
	val=quotient;
        Seq[HANDS-2-i]=new IncrSequence(CARDS,Index);
	i++;
    }

    Seq[0]=new IncrSequence(CARDS,val);

    int hand,index;

    for (index=0; index<HANDS*CARDS; whom[index++]=HANDS-1);

    for (hand=0; hand<HANDS-1; hand++) {
	int i=0, location=0;
        for (index=0; index<HANDS*CARDS; index++) {
	    if (whom[index]==HANDS-1) {
		if (i<CARDS && location==(*Seq[hand])[i]) {
	            whom[index]=hand;
		    i++;
		}
		location++;
	    }
        }
	delete Seq[hand];
    }
}

template <int CARDS, int HANDS>
ostream &tDeal<CARDS, HANDS>::write(ostream &out) const {
    int index;
    for (index=0; index<HANDS*CARDS; index++) {
	out << (char) (whom[index]+'0');
    }
    return out;
}

template <int CARDS, int HANDS>
ostream &tDeal<CARDS, HANDS>::writelong(ostream &out,
	const char *suit_chars,const char *card_chars) const {
  const char *suit_strings[]={
	"<img src=gifs/S.gif alt=\"S : \">",
	"<img src=gifs/H.gif alt=\"H : \">",
	"<img src=gifs/D.gif alt=\"D : \">",
	"<img src=gifs/C.gif alt=\"C : \">"
  };
	
  //static char suit_chars[]="SHDC";
  //static char card_chars[]="AKQJT98765432";

  char a[HANDS][HANDS][CARDS+2];
  char *p[HANDS][HANDS];
  char *result,*rp;
  int suit,hand,card,where;

  for(hand=0; hand<4; hand++)
    for(suit=0; suit<4; suit++)
      p[hand][suit]=&a[hand][suit][0];

  for (suit=0; suit<HANDS; suit++)
    for (card=0; card <CARDS; card += 1) {
	*(p[whom[suit*CARDS+card]][suit]++)=card_chars[card % CARDS];
    }

  for(hand=0; hand<HANDS; hand++)
    for(suit=0; suit<HANDS; suit++)
      if (p[hand][suit] == &a[hand][suit][0]) {
	strcpy(p[hand][suit],"---");
      } else {
	*(p[hand][suit])=0;
      }

  rp=result=new char [2048];

  for (suit=0; suit<HANDS; suit++)  {
    sprintf(rp,"          %s %s\n",
		suit_strings[suit], a[0][suit]);
    rp=rp+strlen(rp);
  }
  *rp++ = '\n';

  for (suit=0; suit<HANDS; suit++) {
    char suitc=suit_chars[suit];
    sprintf(rp," %s %-13s  %s %-13s\n",suit_strings[suit],a[3][suit],
	       suit_strings[suit],a[1][suit]);
    rp=rp+strlen(rp);
  }
  *rp++ = '\n';

  for (suit=0; suit<HANDS; suit++) {
    sprintf(rp,"          %s %s\n",suit_strings[suit],a[2][suit]);
    rp=rp+strlen(rp);
  }

  // remove last newline
  *(rp-1)=0;

  out << result;
  delete [] result;
  return out;

}
