//
// Copyright 1996, Thomas Andrews (thomaso@best.com)
// Free for non-commercial use
//
#ifndef __DEAL__H__
#define __DEAL__H_
#include "stdio.h"
#include "IncrSeq.H"

//
// The obvious first question about this code is: "Why templates?"
//
// There were two reasons for this choice:
//
//      1) To gain more experience with templates, particularly
//      with g++, where I had no prior experience.
//
//      2) Efficiency.  I wanted to write the code generically,
//      so that I could have any number of "hands" and any number
//      of cards per hand.  Using templates, the code optimizes
//      better once the parameters are put in place.  For instance,
//      the loop:
//
//            for (i=0; i<CARDS; i++) foo(i);
//
//      can be reduced by the optimizer to:
//
//	      foo(0); foo(1); foo(2); foo(3);
//
//      with CARDS fixed at compile time rather than at runtime.
//      Other frequent computations include CARDS*HANDS, which
//      would be a runtime constant using templates.
//
//	Probably only a slight efficiency improvement,
//      but one I wanted to make, mostly as an experiment.
//
//  In general, we will do something like:
//
//      typedef tDeal<13,4> Deal;
//  
//  Near the top of our program to set the deal parameters.
//
//  Another interesting option would have been to allow different-sized
//  hands - that is, hand A gets 12, hand B gets 10 cards, hand C
//  gets 11 cards.  The same algorithm would work, I suppose, but I
//  like the basic example - we have just enough cards in a deck
//  to deal an equal number of cards to every person at the table.
//  It codes just slightly more easily, and templates wouldn't have
//  worked there.
//

template <int CARDS,int HANDS>
class tDeal {
private:
  int whom[CARDS*HANDS];
public:
  tDeal(const int *deal_array);
  tDeal(const char *string);

	// Used conversion operators to do the
        // "encode" and "decode" functionality.
        // Not wise choice - such conversion
	// operators should be reserved for "obvious"
	// conversion functions.  Chose to do it this way for
	// further practice using this C++ feature,
	// and to emphasize the identification between
	// "page numbers" and the "deal."
        /////////////
  operator Integer();  // Encode
  tDeal(Integer);      // Decode

  int towhom(int);
  ostream &write(ostream &) const;
  ostream &writelong(ostream &,
	const char *sn="SHDC", const char *cn="AKQJT98765432") const;
};

#include "Deal.C"
#endif
